/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// This is a SQL serializer plugin. It saves and restores the msessiond 
// state at startup, shutdown, and optionally during operation.

#include <stdio.h>
#include "plugapi.h"


// Used to call back into PostgreSQL
MSESSION_EXPORTS *msessiond;

/* Saves a session into the msession_sessions table */
static void savesession(MSql *conn, int serid, char *session,char *value, int timeout)
{
	static char szfmt[]=
		"insert into msession_sessions (serid,session,value,timeout) values (%d,'%s','%s',%d);";
	conn->Execf(szfmt, serid, session, value, timeout);
}

/* Saves session values to the msession_values table */
static void savesessionvalue(MSql *conn, int serid, char *session, char * name, char *value)
{
	static char szfmt[]=
		"insert into msession_values (serid,session,name,value) values (%d,'%s','%s','%s');";
	conn->Execf(szfmt, serid, session, name, value);
}

/* Get session values from the database */
static void getsessionvalues(MSql *conn, REQB **reqb, int serid)
{
	char szfmt[]="select session, name, value from msession_values where serid = %d order by session";
	void *res = conn->ExecfResult(szfmt, serid);

	if(res)
	{
		int n = conn->RowsCount(res);

		for(int i=0; i < n; i++)
		{
			char *session = conn->ColumnValue(res,i, 0);
			char *name = conn->ColumnValue(res,i, 1);
			char *value = conn->ColumnValue(res,i, 2);
			if(session && name && value)
				msessiond->ExecuteRequest(reqb, MREQ_SETVAL, session, name, value, 0);
		}
		conn->FreeResult(res);
	}
}

/* Get sessions from the database */
static void getsessions(MSql *conn, REQB **reqb,int serid)
{
	char szfmt[]="select session, value, timeout from msession_sessions where serid = %d";

	void *res = conn->ExecfResult(szfmt, serid);

	if(res)
	{
		int n = conn->RowsCount(res);

		for(int i=0; i < n; i++)
		{
			char *session = conn->ColumnValue(res,i, 0);
			char *value = conn->ColumnValue(res,i, 1);
			char *timeout = conn->ColumnValue(res,i, 2);
			if(session)
				msessiond->ExecuteRequest(reqb, MREQ_CREATE, session,"",value,atoi(timeout));
		}
		conn->FreeResult(res);
	}
}

// The PLUGIN_RESTORE routine is called at startup to restore the state of the
// msessiond daemon.
static int PLUGIN_RESTORE(void *p1, void *p2)
{
	int serid=-1;
	REQB **preq = (REQB **) p1;
	msessiond->errprintf("SQL Serializer plugin restore begin");

	MSql *conn= MSql::Create( (*preq)->req.datum);

	if(!conn)
	{
		(*preq)->req.stat = REQ_ERR;
		(*preq)->req.param = REQE_UNKNOWN;
		return (*preq)->req.stat;
	}

	if(!conn->ExecInt("select serid from msession_status", &serid))
	{
		delete conn;
		(*preq)->req.stat = REQ_ERR;
		(*preq)->req.param = REQE_UNKNOWN;
		return (*preq)->req.stat;
	}
	else
	{
		getsessions(conn, preq,serid);
		getsessionvalues(conn, preq,serid);
		(*preq)->req.stat = REQ_OK;
	}

	delete conn;

	msessiond->errprintf("SQL Serializer plugin restore end");
	return (*preq)->req.stat;
}

// Called for each value/pair MNode in a session
// p1 is the SERIALIZE_PARAM structure
// p2 has session name
static Boolean sessionCallBack(MNode *mn, void *p1, void *p2)
{
	SERIALIZE_PARAM *psp = (SERIALIZE_PARAM *) p1;
	
	char * name = (char *)mn->KeyGet();
	char * value = (char *)mn->DataGet();

	savesessionvalue((MSql *)psp->vp_var1, (int)psp->n_var1, (char *)p2, name, value) ;
	return TRUE;
}

// The PLUGIN_SERIALIZE routine is called at shutdown, and optionally during
// operation, to save the current state of the msession daemon.
// In the SERIALIZE_PARAM structure:
// n_var1 is the current serialization id
// n_var2 is the previous serialization id
// vp_var1 is the PostgreSQL connection
// vp_var2 is unused.
static int PLUGIN_SERIALIZE(void *p1, void *p2)
{
	SERIALIZE_PARAM *psp = (SERIALIZE_PARAM *) p2;
	REQB **preq = (REQB **) p1;
	MSql *conn = (MSql *)psp->vp_var1;

	if(psp->cmd == PLUGIN_SERIALIZE_BEGIN)
	{
		// Called first to initialize the plugin's serialization function.
		// In this case, setup the database.

		conn = MSql::Create((*preq)->req.datum);

		if(!conn)
		{
			// Return PostgreSQL's error
			(*preq)->req.len = strlen((*preq)->req.datum)+1;
			return FALSE;
		}
		psp->vp_var1 = conn;

		psp->n_var2 = 0;

		msessiond->errprintf("%s SQL serialize", timestr(0));

		// Get the last serialization number
		conn->ExecInt("select serid from msession_status", &psp->n_var2);

		// Begin a transaction to keep it atomic
		conn->Begin();

		// Get the next serialization number
		if(conn->ExecInt("select nextval('msession_serial')", &psp->n_var1))
		{
			conn->Execf("update msession_status set serid = %d, logtime = now()", psp->n_var1);
			return TRUE;
		}
		else
		{
			// Couldn't get it going, about and stop
			conn->Abort();
			delete conn;
			(*preq)->req.stat = REQ_ERR;
			(*preq)->req.param = REQE_UNKNOWN;
			return FALSE;
		}
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_END)
	{
		// Called last to finish the plugin's serialization function.
		// In this case, commit the transaction, and if all is well, delete the old data.
		if(conn->Commit() && psp->n_var2)
		{
			// delete old values
			conn->Exec("insert into msession_log select * from msession_status");
			conn->Execf("delete from msession_values where serid = %d", psp->n_var2);
			conn->Execf("delete from msession_sessions where serid = %d", psp->n_var2);
			(*preq)->req.stat = REQ_OK;
		}
		else
		{
			(*preq)->req.stat = REQ_ERR;
			(*preq)->req.param = REQE_UNKNOWN;
		}
		delete conn;
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_SESSION)
	{
		// Called for each valid session in the daemon. It is the job of the
		// plugin to enumerate and save all the variables in a session.

		assert(psp->session);
		int timeout = (psp->session->TestFlag(MSESS_FIXED)) ? -1 : psp->session->GetTimeout();

		char *session = (char *) psp->session->KeyGet();
		char *data = (char *) psp->session->DataGet();

		if(!data) 
			data = "";
		savesession(conn, psp->n_var1, session, data, timeout);
		psp->session->Enum(sessionCallBack, (void *) psp, (void *)session);
	}
	else
	{
		(*preq)->req.stat = REQ_ERR;
		(*preq)->req.param = REQE_UNKNOWN;
		if(conn)
			delete conn;
		msessiond->errprintf("Unknown serialize command %d", psp->cmd);
		return FALSE;
	}
	return TRUE;
}

FN_HANDLER(default)
{
	return MPLUGIN_VER;
}

FN_INIT(default)
{
	PLUGFN_OPEN(entries)
	PLUGFN_DECL(PLUGIN_SERIALIZE)
	PLUGFN_DECL(PLUGIN_RESTORE)
	PLUGFN_CLOSE

	msessiond = (MSESSION_EXPORTS *) exports;
	*imports = (PLUGFN *)entries;

	msessiond->errprintf("SQL Serializer plugin loaded");
	return NULL;
}

FN_SHUTDOWN(default)
{
}
