/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
/**
This is not a strict metaphone algorithm.
It is tailored to be a little bit less fuzzy.
Most sounds are simply their letter.
Except:
X sounds like KS
Q sounds like KW

Substitutions:
Unused sound/letters are used to represent unique sounds which
are produced by a combination of letters.
'X' is the "sh" sound.
'Q' is the "th" sound.
'E' is the long 'E' or long 'I' sound at the ends of words.

This algorithm is used for searching a dictionary for phrases
it is not a spelling algorithm.
The subtle difference is that a spelling dictionary would suggest
possible spellings, and a user would choose. This algorithm is
used to index words in a group of documents and again used in
their retreival. It needs to be able to differenciate words well 
enough, but also handle minor mispellings.
In this application, "good enough" may be better than "great."

When END_VOWELS is defined
Words which end with 'a' have an 'e' at the end
end in 'i' have an 'e' at the end
end in 'ie' have an 'e' at the end
end in 'o' have an 'o' at the end
end in 'u' have a 'o' at the end
end with 'y' have an 'e' at th end
There are few words which end in 'e' which have an 'e' sound.

When ENCODE_VOWELS is defined
An 'a' or 'e' or 'i' are represented as 'e'
An 'o' or 'u' sound is represented as 'o'
*/

#define IS_AE			(c=='a' && cn=='e')
#define IS_WR			(c=='w' && cn=='r') 
#define IS_WH			(c=='w' && cn=='h')
#define IS_GN_KN_PN(c,cn)	(cn == 'n' && (c == 'g' || c == 'n' || c == 'p'))

int METAPHONE_NAME(char *dest, char *str)
{
	// Search for non alpha/numeric chars passed at the beginning
	// Usually stuff for the search engine

	char *dest_save = dest;
	char *work = str;

	while(!ISALNUM(*((unsigned char *)work)))
		*dest++ = *work++;

	char c = work[0];
	char cn = work[1];
	//
	// Handle start of words differently
	//
	if(IS_GN_KN_PN(c,cn) || IS_WR)
	{	// First char is silent, use second char
		work++;
	}
	else if(c == 'x')
	{	// X at beginning of word like xenon, not a 'ks'
		*work = 's';
	}
	else if(IS_WH)
	{	// wh at beginning of word, sounds like 'w'
		work++;
		*work = 'w';
	}

	int n 		= 0;			// count
	char cp 	= 0;			// 'cp'  char previous
	cn		= work[0];
	char cnn 	= work[1];		// 'cnn' char next next

	unsigned mp  = 0;			// meta info
	unsigned m   = 0;			// meta info
	unsigned mn  = META(cn);		
	unsigned mnn = META(cnn);

	while(*work)
	{
		
		cp	= c;		// Old character becomes previous char.
		c 	= cn;		// Current character
		cn 	= cnn;		// Next character
		cnn 	= work[2];	// Next, Next character
		mp 	= m;		// Previous meta info moved to previous	
		m	= mn;		// previous next, moved to current
		mn	= mnn;		// previous 'next next' moved to next
		mnn 	= META(cnn);	// Get 'next next' meta

		work ++;		// Advance to next char
		*dest = c;		// Put 'c' here as default;

		switch(c)
		{
			// Vowels are usually silent.
			// At end of word, an 'a' sound is pretty important.
			// Humor: In boston, 'a' at end of word sounds like 'r'
			// like 'china' pronounced as 'chiner'
			case 'a':
#ifdef ENCODE_VOWELS
				if(!M_ESOUND(mn) && !M_OY(mn))
					*dest++ = 'e';
#elif defined (END_VOWELS)
				if(!cn)
					*dest++ = 'e';
#endif
				break;
		
			// Silent if two in a row, or after an m, as in bubble and bomb
			case 'b': 	
				if(!(cn == 0 && n && cp == 'm') && (cn != 'b')) 
					dest++;
				break;
		
			// Use 'x' for 'cia' or 'ch'  as in 'caio' or 'cheat'
			// Use 's' for 'ci' or 'ce' or 'cy' as in 'ceil' etc.
			// SILENT if 'sci' or 'sce' or 'scy' (the 's' sound will come from the 's' char.
			// Sounds like 'k' otherwise.
			case 'c': 	
				*dest = 'k';
				if((cp == 's') && M_IEY(mn))
					break;
				else if(M_IEY(mn))
					*dest = 's';
				else if(cn == 'h' || (cn=='i' && cnn =='a'))
					*dest = 'x';
				if((*dest == 'k') && (cn == c))
					break;
				dest++;
				break;

			// SILENT dge, dgi, dgy as in 'dodge'
			// Sounds like 'd' otherwise 
			case 'd': 	
				if(!(cn == 'g' && M_IEY(mnn)))
					*dest++ = 'd';
				break;

			// Vowels are usually silent.
			case 'e':
#ifdef ENCODE_VOWELS
				if(cn && !M_ESOUND(mn) && cn != 'y')
					*dest++='e';
#endif
				break;

			// f sounds like f
			case 'f':
				if(cn != 'f')
					*dest++='f';
				break;

			// SILENT if 'gh' 'gn' and not at the end of a word or before a vowel.
			// Sounds like 'j' if gi, ge, or gy
			// SILENT if after 'n' because nothin' is is often used for nothing.
			// Sounds like 'g' otherwise
			case 'g': 	
				if(M_GN(mn))
					break;
				if((cp != 'g') && M_IEY(mn))
					*dest++ = 'j';
				else if(cp == 'n' && !cn)
					break;
				else if(cn != 'g')
					*dest++ = 'g';
				break;	

			// Silent if ch,sh,ph,th,gh or after a vowel, and no vowels follow
			// Silent if follows an 'r' as in rhapsody
			case 'h': 	
				if(!(M_VARSON(mp) || (M_VOWEL(mp) && !M_VOWEL(mn)) || (cp=='r'))) 
					*dest++='h';
				break;

			// Vowels are silent, but an 'ie' or 'i' at the end of a word
			// sounds like 'e' if we use END_VOWELS
			case 'i':	
#ifdef ENCODE_VOWELS
				if(cn == 'e' && !cnn)
					*dest++ = 'e';
				else if((!cn || !M_ESOUND(mn)) && cn != 'y')
					*dest++ = 'e';
#elif defined (END_VOWELS)
				if(!cn || (cn == 'e' && !cnn))
					*dest++ = 'e';
#endif
				break;
			// j sounds like j
			case 'j':
				if(cn != 'j')
					*dest++='j';
				break;

			// Silent if after a 'c' or before an 'n' or 'k'
			case 'k': 	
				if((cp != 'c') && !M_NK(mn))
					*dest++='k';
				break;

			// Silent if before another 'l' as in lilly
			case 'l': 	
				if(cn != 'l') 
					*dest++='l';
				break;

			// Sounds like m
			// SILENT in mutiples, like momma
			case 'm': 	
				if(cn != 'm') 
					*dest++ = 'm';
				break;

			// Silent if before another 'n' as in connie
			case 'n':	
				if(cn != 'n') 
					*dest++='n';
				break;

			// Vowels are silent.
			// Except that at the end of a word, is important
			case 'o':	
#ifdef ENCODE_VOWELS
				if((!cn || !M_OSOUND(mn)))
					*dest++ = 'o';	
#elif defined (END_VOWELS)
				if(!cn || (cn == 'h' && cnn == 0)) 
					*dest++ = 'o';	
#endif
				break;

			// Sounds life 'f' if an h follows, and silent if another p follows
			// as in hippo
			case 'p': 	
				if(cn == 'h')
					*dest++ = 'f';
				else if(cn != 'p')
					*dest++ = 'p';
				break;

			// Sounds like 'k' 'w'
			case 'q':	
				if(cn != 'q')
				{
					*dest++ = 'k';
					*dest++ = 'w';
				}
				break;

			// silent if before another r, as in carrer
			// Humor: In boston, 'r' at end of word sounds like 'a'
			// like 'car' pronounced as 'ca'
			case 'r':	
				if(cn != 'r') 
					*dest++ = 'r';
				break;

	
			// Sounds like x (sh) before an h or before sio or sia
			// SILENT is the next char is also an 's' as in success
			case 's': 	
				if(cn == 'h' || (cn == 'i' && M_AO(mnn))) 
					*dest++='x';
				else if(cn != 's')
					*dest++='s';
				break;

			// Sounds like x (sh) if before an 'ia' or 'io'
			// Sounds like o (th) if before an 'h'
			// SILENT if before 'tch'
			// Otherwise it sounds like t.
			// Make sure, two t's make one sound.
			case 't': 	
				if(cn == 'i' && M_AO(mnn))
					*dest++ = 'x';
				else if(cn == 'h')
					*dest++ = 'q';
				else if(! ((cn == 'c' && cnn=='h')||(cn == 't')))
					*dest++ = 't';
				break;

			// Vowels are usually silent
			case 'u':
#ifdef ENCODE_VOWELS
				if((!cn || !M_OSOUND(mn)))
					*dest++ = 'o';	
#elif defined (END_VOWELS)
				if(!cn)
					*dest++='o';
#endif
				break;

			// Metaphone says to use f, but I think it sounds like an 'v'
			case 'v': 	
				if(cn != 'v')
					*dest++='v';
				break;

			// Silent if not followed by a vowel
			case 'w': 	
				if(M_VOWEL(mn) || ((cp == 0) && (cn == 0)) || (cn=='y')) 
					*dest++ = 'w';
				break;

			// Sounds like 'k' 's'
			case 'x': 	
				if(cn != 'x')
				{
					*dest++='k';
					*dest++='s';
				}
				break;

			// Vowels are aeiou and sometimes y
			// Treated as a 'y' if not followed by a vowel, 
			// as in as in yellow and toyota.
			// Treated as long 'e' sound at end of word.
			// Otherise, if it is not at the end of the word
			// It is a vowel, as in analyse or allyssa. 
			
			case 'y': 	
#ifdef END_VOWELS
				if(!cn)
					*dest++='e';
#elif defined(ENCODE_VOWELS)
				if(!M_VOWEL(mn))
					*dest++='e';
				else if(M_VOWEL(mn))
					*dest++='y';
#endif
				break;

			// 'Z' sounds like 's'
			case 'z': 	
				if(cn != 'z')
					*dest++ = 's';
				break;

			case	'0':
			case	'1':
			case	'2':
			case	'3':
			case	'4':
			case	'5':
			case	'6':
			case	'7':
			case	'8':
			case	'9':
				*dest++ = c;
				break;
				
			// Default is to ignore it.
			default: 	
				break;
		}
		*dest=0;
		n++;
	}
	return dest - dest_save;
}
