/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 

#include "Phoenix.h"


#ifdef WIN32
#define MUTEX_DEFINED
// Win32 version of "in-process" mutexes. No kernel transition
// need be made if mutex can be locked.
#ifndef INLINE_MUTEX
void MutexLock(MUTEX *mutex)
{
	EnterCriticalSection(mutex);
}
void MutexUnlock(MUTEX *mutex)
{
	LeaveCriticalSection(mutex);
}
#endif

void MutexInit(MUTEX *mutex)
{
	InitializeCriticalSection(mutex);
}

void MutexDestroy(MUTEX *mutex)
{
	DeleteCriticalSection(mutex);
}

void EventWait(EVENT *event, MUTEX *mutex, int useconds)
{
	MutexUnlock(mutex);
	WaitForSingleObject(*event, useconds*10);
	MutexLock(mutex);
}

void EventInit(EVENT *event)
{
	*event = CreateEvent(NULL,FALSE,TRUE,NULL);
}
void EventDestroy(EVENT *event)
{
	CloseHandle(*event);
}
void EventSignal(EVENT *event)
{
	SetEvent(*event);
}
#endif

#ifdef POSIX
#define MUTEX_DEFINED
#ifndef INLINE_MUTEX
void MutexLock(MUTEX *mutex)
{
	pthread_mutex_lock(mutex);
}
void MutexUnlock(MUTEX *mutex)
{
	pthread_mutex_unlock(mutex);
}
#endif
void MutexInit(MUTEX *mutex)
{
	pthread_mutex_init(mutex, NULL);
}
void MutexDestroy(MUTEX *mutex)
{
	pthread_mutex_destroy(mutex);
}
void EventWait(EVENT *event, MUTEX *mutex, int useconds)
{
	pthread_cond_wait(event, mutex);
}
void EventInit(EVENT *event)
{
	pthread_cond_init(event, NULL);
}
void EventDestroy(EVENT *event)
{
	pthread_cond_destroy(event);
}
void EventSignal(EVENT *event)
{
	pthread_cond_signal(event);
}
#endif

MMutex::MMutex()
{
	MutexInit(&m_mutex);
	EventInit(&m_event);
}
MMutex::~MMutex()
{
	EventDestroy(&m_event);
	MutexDestroy(&m_mutex);
}
void MMutex::wait(int seconds, int usec)
{
	EventWait(&m_event, &m_mutex, usec);
}
void MMutex::signal()
{
	EventSignal(&m_event);
}

int MMutex::inc(volatile int *var)
{
#ifdef USE_ATOM_OPS
	atom_inc(var);
#else
	MutexLock(&m_mutex);
	++(*var);
	MutexUnlock(&m_mutex);
#endif
	return *var;
}
int MMutex::dec(volatile int *var)
{
#ifdef USE_ATOM_OPS
	atom_dec(var);
#else
	MutexLock(&m_mutex);
	 --(*var);
	MutexUnlock(&m_mutex);
#endif
	return *var;
}

int MMutex::xchg(volatile int *m, int intval)
{
#ifdef USE_ATOM_OPS
	return atom_xchg(m, intval);
#else
	MutexLock(&m_mutex);
	int n = *m;
	*m = intval;
	MutexUnlock(&m_mutex);
	return n;
#endif
}
int MMutex::add(volatile int *m, int num)
{
#ifdef USE_ATOM_OPS
	atom_add(m, num);
#else
	MutexLock(&m_mutex);
	*m += num;
	MutexUnlock(&m_mutex);
#endif
	return *m;
}
int MMutex::sub(volatile int *m, int num)
{
#ifdef USE_ATOM_OPS
	atom_sub(m, num);
#else
	MutexLock(&m_mutex);
	*m -= num;
	MutexUnlock(&m_mutex);
#endif
	return *m;
}

