/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// #define TRACE_DEBUG 
#include "session.h"


static int max_locktime = 30;

// This is how much we pad by.
#define PAD(l)	(((l)+8)&(~7))
// This routine checks to see if the new pointer will fit
// in the old memory, if not, a new pointer will be allocated.
// This may save a malloc and and there, which could be size able.

#define CHKALLOC(len, newlen, data)\
	(len > (unsigned)newlen && data) ? data : chkalloc(&len, newlen, data)

void * chkalloc(unsigned int *plen, unsigned int len, void *data)
{
	if(data)
	{
		if(*plen < len)		
		{
			phmfree(data);
			*plen = PAD(len);
			data = (char *) phmalloc(*plen);
		}
	}
	else
	{
		*plen = PAD(len);
		data = (char *) phmalloc(*plen);
	}
	assert(data);
	assert(*plen >= len);
	return data;
}
Session::Session(SessionMgr *smgr)
{
	TRACE_FUNCT("Session:Session");
	m_tcreate = time(0);
	m_tlast = m_tcreate;
	m_locked = 0;
	m_flags = 0;
	m_data = 0;
	m_timeout=0;
	m_lendata=0;
}
Session::~Session()
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	TRACE_FUNCT("Session:~Session");
	DeleteAll(MNodeDataDel,NULL,NULL);
	if(m_data)
		phmfree(m_data);
}
int Session::CompareValue(char *name, char *value)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	TRACE_FUNCT("Session:CompareValue");
	MNode *node = FindStr(name);

	if(node)
	{
		char *p = (char *) node->DataGet();
		TRACE_STR("Found:");
		TRACE_STR(p);
		return strcmp(value,p);
	}
	else
		return -1;
}
int Session::GetValue(char *name, char *data, int *datasize, Boolean updateTime)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	TRACE_FUNCT("Session:GetValue");
	MNode *node = FindStr(name);
	int len = 0;

	if(updateTime)
		m_tlast = time(0);

	if(node)
	{
		char *p = (char *) node->DataGet();
		len = strlen(p)+1;
		if((len <= *datasize) && data)
		{
			memcpy(data,p,len);
			TRACE_STR("Found:");
			TRACE_STR(p);
		}
		else
		{
			*datasize = len;
			return -1;
		}
	}
	else
	{
		TRACE_STR("Not found");
		TRACE_STR(name);
	}
	return len;
}
void Session::SetValue(char *name, char *value)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	TRACE_FUNCT("Session:SetValue");
	m_tlast = time(0);
	MNode *node = FindStr(name);

	int len = strlen(value)+1;

	if(!node)
	{
		node = new (name) MNode();
		Insert(node);
	}

	node->m_data = CHKALLOC(node->m_prop, len, node->m_data);
	memcpy((char *)node->m_data, value,len);

#ifdef TRACE_DEBUG
	mferr.mfprintf("Set: %s = %s\n", name, node->m_data);
#endif
}
void Session::SetTimeout(time_t timeout)
{
	m_timeout = timeout;
}
time_t Session::GetTimeout(void)
{
	return m_timeout;
}
time_t Session::GetLastTime(void)
{
	return m_tlast;
}
time_t Session::GetCreateTime(void)
{
	return m_tcreate;
}
void Session::SetLast(void)
{
	m_tlast = time(0);
}
time_t Session::GetTTL(time_t now, time_t default_age)
{
	int age = now - m_tlast;
	return (m_timeout) ? m_timeout - age : default_age - age;
}

Boolean Session::HasExpired(time_t now, time_t default_age)
{
	int age = (m_timeout) ? m_timeout : default_age;
	return ((now - m_tlast) > age) ? TRUE : FALSE;
}
unsigned int Session::SetFlag(unsigned int flag)
{
	m_tlast = time(0);
	m_flags |= flag;
	return m_flags;
}
unsigned int Session::ResetFlag(unsigned int flag)
{
	m_tlast = time(0);
	m_flags &= ~flag;
	return m_flags;
}
unsigned int Session::GetFlags()
{
	return m_flags;
}
Boolean Session::TestFlag(unsigned int flag)
{
	return (m_flags & flag) == flag;
}
void * Session::operator new (size_t size, char * name)
{
	int cbname = strlen(name)+1;
	int cb = size + cbname;
	char * p = (char *) phmalloc(cb);
	if(p)
	{
		Session *s = (Session *)p;
		s->m_key = (char *)&p[size];
		strcpy((char *)s->m_key, name);
	}
	return (void *) p;
}
int Session::lock(void)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	time_t now = time(0);
	int locked = 0;
	m_tlast = now;

	if(m_locked)
	{
		if((now - m_locked) >= max_locktime)
			locked = m_locked = now;
	}
	else
	{
		locked = m_locked = now;
	}
	return locked;
}
Boolean Session::unlock(int key)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	m_tlast = time(0);
	if(key == m_locked)
	{
		m_locked = 0;
		return TRUE;
	}
	else
		return FALSE;
}

Boolean Session::islocked()
{
	return (m_locked) ? TRUE : FALSE;
}

void Session::DeleteSessionData(char *szfilePath)
{
	if(szfilePath)
	{
		char buffer[PATH_MAX];
		sprintf(buffer, "%s/msession_%s", szfilePath, (char *)m_key);
		remove(buffer);
	}
}
char *Session::SetSessionData(char *value, char *szfilePath,int len)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	TRACE_FUNCT("Session:SetSessionData");
	m_tlast = time(0);

	if(!len)
		len = strlen(value)+1;

	if(szfilePath)
	{
		MIoFile mf;
		char buffer[PATH_MAX];
		sprintf(buffer, "%s/msession_%s", szfilePath, (char *)m_key);

		if(mf.open(buffer, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR))
		{
			mf.write(value, len);
			return value;
		}
		else
		{
			mferr.mfprintf("Open for write failed %s\n", buffer);
			return NULL;
		}
	}
	else
	{
		m_data = CHKALLOC(m_prop, len, m_data);
		memcpy((char*)m_data, value, len);
		m_lendata = len;

		TRACE_FUNCT(m_data);
		return (char *)m_data;
	}
}
char *Session::GetSessionDataReference(int *len)
{
	*len = m_lendata;
	return (char *)m_data;
}
int Session::GetSessionData(char *data, int *datasize, char *szfilePath)
{
	TRACE_FUNCT("Session:GetSessionData");
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	m_tlast = time(0);
	if(szfilePath)
	{
		MIoFile mf;
		char buffer[PATH_MAX];
		sprintf(buffer, "%s/msession_%s", szfilePath, (char *)m_key);
		if(mf.open(buffer, O_RDWR, S_IRUSR|S_IWUSR))
		{
			struct stat buf;
			if(mf.stat(&buf))
			{
				if(buf.st_size <= *datasize)
				{
					TRACE_STR("Buffer big enough");
					mf.read(data, buf.st_size);
					return buf.st_size;
				}
				else
				{
					TRACE_STR("Buffer too small");
					*datasize = buf.st_size;
					return -1;
				}
			}
		}
	}
	else
	{
		if(m_data)
		{
			register int len = m_lendata;
			TRACE_STR("Have Data");
			char *p = (char *) m_data;
			TRACE_VAL(strlen(p)+1);
			TRACE_VAL(m_lendata);
			if(len <= *datasize)
			{
				TRACE_STR("Buffer big enough");
				memcpy(data,p,len);
				return len;
			}
			else
			{
				TRACE_STR("Buffer too small");
				*datasize = len;
				return -1;
			}
		}
	}
	return 0;
}
int Session::nextval(char *name, char *data)
{
#ifdef SESSION_MUTEX
	AutoMutex lock(&m_mtx);
#endif
	TRACE_FUNCT("Session:nextval");
	m_tlast = time(0);
	MNode *node = FindStr(name);

	int count = 0;

	if(!node)
	{
		node = new (name) MNode();
		Insert(node);
	}
	else
	{
		char *szcount = (char *)node->DataGet();
		if(szcount)
			count = atoi(szcount);
	}
	count++;
	_itoa(count,data,10);
	unsigned int len = strlen(data)+1;
	node->m_data = CHKALLOC(node->m_prop, len, node->m_data);
	memcpy(node->m_data, data, len);

#ifdef TRACE_DEBUG
	mferr.mfprintf("nextval: %s = %s[%d]\n", name, data, count);
#endif
	return count;
}

///////////////////////// Session manager code ////////////////////////

SessionMgr::SessionMgr(int size) : MHash(size)
{
	TRACE_FUNCT("SessionMgr:SessionMgr");
}

void SessionMgr::SetMaxLockTime(int t)
{
	max_locktime = t;
}
